@@ -28,6 +28,10 @@
 #include <gdk/gdk.h>
 #include <gdk/gdkx.h>
 
+#ifdef HAVE_XF86MISCSETGRABKEYSSTATE
+# include <X11/extensions/xf86misc.h>
+#endif /* HAVE_XF86MISCSETGRABKEYSSTATE */
+
 #include "gs-window.h"
 #include "gs-grab.h"
 #include "gs-debug.h"
@@ -81,6 +85,55 @@ grab_string (int status)
         }
 }
 
+#ifdef HAVE_XF86MISCSETGRABKEYSSTATE
+/* This function enables and disables the Ctrl-Alt-KP_star and 
+   Ctrl-Alt-KP_slash hot-keys, which (in XFree86 4.2) break any
+   grabs and/or kill the grabbing client.  That would effectively
+   unlock the screen, so we don't like that.
+
+   The Ctrl-Alt-KP_star and Ctrl-Alt-KP_slash hot-keys only exist
+   if AllowDeactivateGrabs and/or AllowClosedownGrabs are turned on
+   in XF86Config.  I believe they are disabled by default.
+
+   This does not affect any other keys (specifically Ctrl-Alt-BS or
+   Ctrl-Alt-F1) but I wish it did.  Maybe it will someday.
+ */
+static void
+xorg_lock_smasher_set_active (gboolean active)
+{
+        int status;
+
+        if (active) {
+                gs_debug ("Enabling the x.org grab smasher");
+        } else {
+                gs_debug ("Disabling the x.org grab smasher");
+        }
+
+        gdk_error_trap_push ();
+
+        status = XF86MiscSetGrabKeysState (GDK_DISPLAY (), active);
+
+        gdk_error_trap_pop ();
+
+        if (active && status == MiscExtGrabStateAlready) {
+                /* shut up, consider this success */
+                status = MiscExtGrabStateSuccess;
+        }
+
+        gs_debug ("XF86MiscSetGrabKeysState(%s) returned %s\n",
+                  active ? "on" : "off",
+                  (status == MiscExtGrabStateSuccess ? "MiscExtGrabStateSuccess" :
+                   status == MiscExtGrabStateLocked  ? "MiscExtGrabStateLocked"  :
+                   status == MiscExtGrabStateAlready ? "MiscExtGrabStateAlready" :
+                   "unknown value"));
+}
+#else
+static void
+xorg_lock_smasher_set_active (gboolean active)
+{
+}
+#endif /* HAVE_XF86MISCSETGRABKEYSSTATE */
+
 void
 gs_grab_window (GdkWindow *window,
                 GdkScreen *screen,
@@ -88,13 +141,17 @@ gs_grab_window (GdkWindow *window,
 {
         gboolean result = FALSE;
 
-        do
-                result = gs_grab_move_keyboard (window, screen);
-        while (!result);
+        xorg_lock_smasher_set_active (FALSE);
 
-        do
+        do {
+                result = gs_grab_move_keyboard (window, screen);
+                gdk_flush ();
+        } while (!result);
+
+        do {
                 result = gs_grab_move_mouse (window, screen, hide_cursor);
-        while (!result);
+                gdk_flush ();
+        } while (!result);
 }
 
 static int
@@ -148,26 +205,38 @@ gs_grab_get_mouse (GdkWindow *window,
         return status;
 }
 
+void
+gs_grab_keyboard_reset (void)
+{
+        keyboard_grab_window = NULL;
+        keyboard_grab_screen = NULL;
+}
+
 static gboolean
 gs_grab_release_keyboard (void)
 {
         gs_debug ("Ungrabbing keyboard");
         gdk_keyboard_ungrab (GDK_CURRENT_TIME);
 
-        keyboard_grab_window = NULL;
-        keyboard_grab_screen = NULL;
+        gs_grab_keyboard_reset ();
 
         return TRUE;
 }
 
-static gboolean
+void
+gs_grab_mouse_reset (void)
+{
+        mouse_grab_window = NULL;
+        mouse_grab_screen = NULL;
+}
+
+gboolean
 gs_grab_release_mouse (void)
 {
         gs_debug ("Ungrabbing pointer");
         gdk_pointer_ungrab (GDK_CURRENT_TIME);
 
-        mouse_grab_window = NULL;
-        mouse_grab_screen = NULL;
+        gs_grab_mouse_reset ();
 
         return TRUE;
 }
@@ -182,17 +251,43 @@ gs_grab_move_mouse (GdkWindow *window,
         GdkScreen *old_screen;
         gboolean   old_hide_cursor;
 
+        /* if the pointer is not grabbed and we have a
+           mouse_grab_window defined then we lost the grab */
+        if (! gdk_pointer_is_grabbed ()) {
+                gs_grab_mouse_reset ();
+        }
+
+        if (mouse_grab_window == window) {
+                gs_debug ("Window %X is already grabbed, skipping",
+                          (guint32) GDK_WINDOW_XID (mouse_grab_window));
+                return TRUE;
+        }
+
+#if 0
+        gs_debug ("Intentionally skipping move pointer grabs");
         /* FIXME: GTK doesn't like having the pointer grabbed */
         return TRUE;
+#else
+        if (mouse_grab_window) {
+                gs_debug ("Moving pointer grab from %X to %X",
+                          (guint32) GDK_WINDOW_XID (mouse_grab_window),
+                          (guint32) GDK_WINDOW_XID (window));
+        } else {
+                gs_debug ("Getting pointer grab on %X",
+                          (guint32) GDK_WINDOW_XID (window));
+        }
+#endif
 
+        gs_debug ("*** doing X server grab");
         gdk_x11_grab_server ();
 
         old_window = mouse_grab_window;
         old_screen = mouse_grab_screen;
         old_hide_cursor = mouse_hide_cursor;
 
-        if (old_window)
+        if (old_window) {
                 gs_grab_release_mouse ();
+        }
 
         result = gs_grab_get_mouse (window, screen, hide_cursor);
 
@@ -206,6 +301,7 @@ gs_grab_move_mouse (GdkWindow *window,
                 gs_grab_get_mouse (old_window, old_screen, old_hide_cursor);
         }
                 
+        gs_debug ("*** releasing X server grab");
         gdk_x11_ungrab_server ();
         gdk_flush ();
 
@@ -220,13 +316,31 @@ gs_grab_move_keyboard (GdkWindow *window,
         GdkWindow *old_window;
         GdkScreen *old_screen;
 
+        if (keyboard_grab_window == window) {
+                gs_debug ("Window %X is already grabbed, skipping",
+                          (guint32) GDK_WINDOW_XID (keyboard_grab_window));
+                return TRUE;
+        }
+
+        if (keyboard_grab_window) {
+                gs_debug ("Moving keyboard grab from %X to %X",
+                          (guint32) GDK_WINDOW_XID (keyboard_grab_window),
+                          (guint32) GDK_WINDOW_XID (window));
+        } else {
+                gs_debug ("Getting keyboard grab on %X",
+                          (guint32) GDK_WINDOW_XID (window));
+
+        }
+
+        gs_debug ("*** doing X server grab");
         gdk_x11_grab_server ();
 
         old_window = keyboard_grab_window;
         old_screen = keyboard_grab_screen;
 
-        if (old_window)
+        if (old_window) {
                 gs_grab_release_keyboard ();
+        }
 
         result = gs_grab_get_keyboard (window, screen);
 
@@ -240,6 +354,7 @@ gs_grab_move_keyboard (GdkWindow *window,
                 gs_grab_get_keyboard (old_window, old_screen);
         }
 
+        gs_debug ("*** releasing X server grab");
         gdk_x11_ungrab_server ();
         gdk_flush ();
 
@@ -266,8 +381,13 @@ gs_grab_nuke_focus (void)
 void
 gs_grab_release_keyboard_and_mouse (void)
 {
+        gs_debug ("Releasing all grabs");
+
         gs_grab_release_mouse ();
         gs_grab_release_keyboard ();
+
+        /* FIXME: is it right to enable this ? */
+        xorg_lock_smasher_set_active (TRUE);
 }
 
 gboolean
@@ -284,8 +404,9 @@ gs_grab_get_keyboard_and_mouse (GdkWindow *window,
 
         for (i = 0; i < retries; i++) {
                 kstatus = gs_grab_get_keyboard (window, screen);
-                if (kstatus == GDK_GRAB_SUCCESS)
+                if (kstatus == GDK_GRAB_SUCCESS) {
                         break;
+                }
 
                 /* else, wait a second and try to grab again. */
                 sleep (1);
@@ -301,19 +422,23 @@ gs_grab_get_keyboard_and_mouse (GdkWindow *window,
 
         for (i = 0; i < retries; i++) {
                 mstatus = gs_grab_get_mouse (window, screen, FALSE);
-                if (mstatus == GDK_GRAB_SUCCESS)
+                if (mstatus == GDK_GRAB_SUCCESS) {
                         break;
+                }
 
                 /* else, wait a second and try to grab again. */
                 sleep (1);
         }
 
-        if (mstatus != GDK_GRAB_SUCCESS)
+        if (mstatus != GDK_GRAB_SUCCESS) {
                 g_warning ("Couldn't grab pointer!  (%s)",
                            grab_string (mstatus));
+        }
 
+#if 0
         /* FIXME: release the pointer grab so GTK will work */
         gs_grab_release_mouse ();
+#endif
 
         /* When should we allow blanking to proceed?  The current theory
            is that a keyboard grab is manditory; a mouse grab is optional.
@@ -326,8 +451,10 @@ gs_grab_get_keyboard_and_mouse (GdkWindow *window,
            activity, so that's not a disaster.
         */
 
-        if (kstatus != GDK_GRAB_SUCCESS)	/* Do not blank without a kbd grab.   */
+        if (kstatus != GDK_GRAB_SUCCESS) {
+                /* Do not blank without a kbd grab.   */
                 return FALSE;
+        }
 
         return TRUE;			/* Grab is good, go ahead and blank.  */
 }
